/*****************************
* Source : pgsql.prg
* System : <unkown>
* Author : Phil Ide
* Created: 02/02/2004
*
* Purpose:
* ----------------------------
* History:
* ----------------------------
*    02/02/2004 14:19 PPI - Created
*****************************/

/*
  For an explanation of how to this works, see:

  C:\Program Files\PostGreSQL\usr\doc\postgresql-7.3.1\html\programmer-client.html

  (assuming you installed to the default directories)
*/

#include "Error.ch"
#include "piCommon.ch"
#include "dll.ch"
#include "dbstruct.ch"
#include "ot4xb.ch"

#xcommand DEFAULT <var1> TO <val1> [,<varN> TO <valN>];
       => IF <var1> == NIL ; <var1> := <val1> ; ENDIF ;
       [; IF <varN> == NIL ; <varN> := <valN> ; ENDIF ]


#define TRUE .T.
#define FALSE .F.

*#pragma Map(PGDeref,"_XBPPGDEREF")

CLASS PGSql
   EXPORTED:
      VAR connID READONLY
      VAR result

      METHOD init
      METHOD connect
      METHOD setdbLogin
      METHOD disconnect
      METHOD reset

      // non-blocking connection
      METHOD connectStart
      METHOD connectPoll
      METHOD resetStart
      METHOD resetPoll

      METHOD setnonblocking
      METHOD isnonblocking
      METHOD sendQuery
      METHOD getResult
      METHOD consumeInput
      METHOD isBusy
      METHOD flush
      METHOD socket
      METHOD requestCancel

      // trace/debug
      METHOD trace
      METHOD untrace


      METHOD exec
      METHOD execParams
      METHOD escapeString
      METHOD escapeBytea
      METHOD unescapeBytea

      METHOD makeEmptyPGResult

      // info
      METHOD db
      METHOD user
      METHOD pass
      METHOD host
      METHOD port
      METHOD options
      METHOD status
      METHOD errorMessage
      METHOD backendPID
ENDCLASS

METHOD PGSql:init()
   ::connID := 0
   return self

METHOD PGSql:setdbLogin(cHost,cPort,cOpts,cTty, cDbName,cUser,cPwd)
   local lOk

   ::connID := PQsetdbLogin(cHost,cPort,cOpts,cTty, cDbName,cUser,cPwd)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connect( cConnString )
   local lOk

   ::connID := PQconnectdb(cConnString)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connectStart( cConnString )
   local lOk

   ::connID := PQconnectdb(cConnString)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connectPoll()
   return PQconnectPoll(::connID)

METHOD PGSql:resetStart()
   return PQresetStart(::connID)

METHOD PGSql:resetPoll()
   return PQresetPoll(::connID)

METHOD PGSql:disconnect()
   PQfinish(::connID)
   return self

METHOD PGSql:reset()
   local lOk

   PQreset(::connID)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:escapeString(c)
   local cBuff := Space( (Len(c)*2)+1 )
   local pError := 0
*
* These forms are deprecated
*
*   PQescapeString( @cBuff, c, Len(c) )

    PQescapeStringConn(::connID, @cBuff, c, Len(c), @pError)

   return cBuff

METHOD PGSql:escapeBytea(cFrom)
*   local cBuff
*   local i := 0
*   cBuff := PQescapeBytea( c, Len(c), @i )
*   return PGDeref(cBuff)

  local cbResult := 0
  local pStr     := 0
  local cStr
  if ( Valtype( cFrom ) $ "CM" )
     pStr := PQescapeByteaConn(::connID,cFrom,Len(cFrom),@cbResult)
     if( pStr != 0 )
        cStr := PeekStr(pStr,0,cbResult) ; PQfreemem( pStr )
        return cStr
     end
  end
return NIL


METHOD PGSql:unescapeBytea(c)
   local cBuff
   local i := 0

   cBuff := PQunescapeBytea( c, Len(c) )
*   return PGDeref(cBuff)
   return(cBuff)



METHOD PGSql:exec(cSQL)
   local xT
   local oQ
   local lOk := TRUE

   ::result := NIL
   xT := PQexec(::connID,cSQL)
   if xT == 0
      lOk := FALSE
   else
      ::result := PGResult():new( self, xT )
   endif
   return lOk

METHOD PGSql:execParams(cSql,nParams,ct,cv,cl,cf,rf)
   local xT
   local oQ
   local lOk := TRUE

   ::result := NIL
   xT := PQexecParams(::connID,cSql,nParams,ct,cv,cl,cf,rf)
   if xT == 0
      lOk := FALSE
   else
      ::result := PGResult():new( self, xT )
   endif
   return lOk


METHOD PGSql:db()
   local c := PQdb(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:user()
   local c := PQuser(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:pass()
   local c := PQpass(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:host()
   local c := PQhost(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:port()
   local c := PQport(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:options()
   local c := PQoptions(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:status()
   return PQstatus(::connID)

METHOD PGSql:errorMessage()
   local c := PQerrorMessage(::connID)
*   c := PGDeref(c)
   return c

METHOD PGSql:backendPID()
   return PQbackendPID(::connID)

METHOD PGSql:makeEmptyPGResult(i)
   local x := PQmakeEmptyPGResult(::connID,i)
   local o

   o := PGResult():new(::connID,x)
   return o


METHOD PGSql:setnonblocking(lNonBlockingMode)
   return PQsetnonblocking(::connID,iif(lNonBlockingMode,1,0)) == 0

METHOD PGSql:isnonblocking()
   return PQisnonblocking(::connID) == 1

METHOD PGSql:sendQuery( cQuery )
   return PQsendQuery(::connID, cQuery) == 1

METHOD PGSql:getResult
   local o := PQgetResult(::connID)
   o := PGResult():new( self, o )
   return o

METHOD PGSql:consumeInput()
   return PQconsumeInput(::connID)

METHOD PGSql:isBusy
   return PQisBusy(::connID) == 1

METHOD PGSql:flush()
   return PQflush(::connID) == 0

METHOD PGSql:socket()
   return PQsocket(::connID) // >0 = connector, <0 is error (-1 no backend!)

METHOD PGSql:requestCancel()
   return PQrequestCancel(::connID) == 1

/*
:requestCancel()
The return value is 1 if the cancel request was successfully dispatched, 0 if not.
(If not, PQerrorMessage tells why not.) Successful dispatch is no guarantee that the
request will have any effect, however. Regardless of the return value of
PQrequestCancel, the application must continue with the normal result-reading
sequence using PQgetResult. If the cancellation is effective, the current command
will terminate early and return an error result. If the cancellation fails (say,
because the backend was already done processing the command), then there will be no
visible result at all.
*/

METHOD PGSql:trace( nFilehandle )
   PQtrace(::connID, nFilehandle )
   return self

METHOD PGSql:untrace()
   PQuntrace(::connID)
   return self



//<!===================================================!>//
CLASS PGResult
   EXPORTED:
      VAR resID READONLY
      VAR curTuple READONLY
      VAR connID
      VAR rows
      VAR cols
      VAR message


      METHOD init
      METHOD destroy
      METHOD clear
      METHOD fname
      METHOD fnumber
      METHOD ftype
      METHOD fmod
      METHOD fsize
      METHOD getValue
      METHOD getLength
      METHOD getIsNULL
      METHOD binaryTuples

      METHOD cmdStatus
      METHOD cmdTuples
      METHOD oidValue
*     METHOD oidStatus // NOT THREAD SAFE
      METHOD resultStatus
      METHOD resStatus
      METHOD resultErrorMessage
      METHOD makeEmptyPGResult

      METHOD escapeString  // changed
      METHOD escapeBytea   // changed
      METHOD unescapeBytea // changed

      ACCESS ASSIGN METHOD NoIvarCallBack

      METHOD skip
      METHOD goTop
      METHOD goTo
      METHOD goBottom
      METHOD eof
      METHOD bof
      METHOD recno
      METHOD position
      METHOD goPosition

      METHOD getFieldNames
      METHOD DataBlock()
      METHOD setUsedFields()

   PROTECTED:
      METHOD NoIVar
      VAR eofMark READONLY
      VAR bofMark READONLY
      VAR UsedField
ENDCLASS

METHOD PGResult:init( oConn, resID )
   ::connID := oConn
   ::resID  := resID
   ::rows := PQntuples(::resID)
   ::cols := PQnfields(::resID)
   ::curTuple := 0
   ::eofMark := (::rows == 0)
   ::bofMark := (::rows == 0)
   ::UsedField := {}
   return self

METHOD PGResult:setUsedFields(aFields)
   IF PCOUNT() > 0
      ::UsedField := aFields
   ENDIF
RETURN ::UsedField

METHOD PGResult:resultStatus()
   return PQresultStatus(::resID)

METHOD PGResult:resStatus(i)
*   return PGDeref(PQresStatus(i))
   return (PQresStatus(i))

METHOD PGResult:makeEmptyPGResult(i)
   return ::connID:makeEmptyPGResult(i)


METHOD PGResult:resultErrorMessage()
*   return PGDeref(PQresultErrorMessage(::resID))
   return (PQresultErrorMessage(::resID))

METHOD PGResult:eof()
   return ::eofMark

METHOD PGResult:bof()
   return ::bofMark

METHOD PGResult:recno()
   return ::curTuple

METHOD PGResult:position()
   local nPC := ::rows/100
   return (::curTuple+1)/nPC

METHOD PGResult:goPosition(n)
   local nPC := ::rows/100

   ::curTuple := int(nPC*n)
   return ::curTuple


METHOD PGResult:skip(n)
   default n to 1
   ::curTuple += n
   do case
      case ::curTuple < 0
         ::curTuple := 0
         ::bofMark := TRUE
         ::eofMark := FALSE

      case ::curTuple > ::rows-1
         ::curTuple := ::rows-1
         ::eofMark := TRUE
         ::bofMark := FALSE

      otherwise
         ::eofMark := FALSE
         ::bofMark := FALSE
   endcase
   return self

METHOD PGResult:goTop()
   ::curTuple := 0
   ::eofMark := FALSE
   ::bofMark := FALSE
   return self


METHOD PGResult:goTo(n)
   if ValType(n) == 'N'
      if n >= 0 .and. n < ::cols
         ::curTuple := n
         ::eofMark := FALSE
         ::bofMark := FALSE

      elseif n < 0
         ::bofMark := TRUE
         ::eofMark := FALSE

      elseif n > ::cols
         ::eofMark := TRUE
         ::bofMark := FALSE

      endif
   endif
   return ::curTuple

METHOD PGResult:goBottom()
   ::curTuple := ::rows-1
   ::eofMark := FALSE
   ::bofMark := FALSE
   return self

METHOD PGResult:getFieldNames()
   local i
   local aRet := Array(::cols)

   for i := 0 to ::cols-1
      aRet[i+1] := ::fname(i)
   next
   return aRet

METHOD PGResult:DataBlock(cFname)
   LOCAL aFields
   LOCAL nPosi   := 0
   local xRet    := ""
   local aClone

   DEFAULT cFname TO ""
*ALTD()
   IF !EMPTY(::UsedField)
      aClone  := ACLONE(::UsedField)
      aFields := AEVAL(aClone,{|a,i| a:= a[DBS_NAME]},,,.T.)
   ELSE
      aFields := ::getFieldNames()
   ENDIF

   nPosi := ASCAN(aFields,{|x| x = cFname} )
   IF nPosi > 0
      xRet := ::getValue(::curTuple, nPosi-1)
   ENDIF
   return xRet


METHOD PGResult:NoIvarCallBack(x)
   local i
   local xRet

   i := PQfnumber(::resID, ::message)
   if i >= 0
      xRet := ::getValue(::curTuple, i)
   else
      ::noIvar(x)
   endif
   return xRet

METHOD PGResult:NoIvar( x )
   local o := Error():new()

   o:args            := x
   o:canDefault      := FALSE
   o:canRetry        := FALSE
   o:canSubstitute   := FALSE
   o:description     := "invalid field name"
   o:operation       := ::className()+":"+::message
   o:severity        := XPP_ES_ERROR
   o:subSystem       := ::className()
   o:thread          := ThreadID()
   Eval( ErrorBlock(), o )
   return self


METHOD PGResult:destroy()
   if !(::resID == NIL)
      PQclear(::resID)
      ::resID := NIL
   endif
   return self

METHOD PGResult:clear()
   ::destroy()
   return self

METHOD PGResult:fname(i)
   local c := PQfname(::resID, i)
*ALTD()
//   return PGDeref(c)
return c

METHOD PGResult:fnumber(cField)
   return PQfnumber(::resID, cField)

METHOD PGResult:ftype(nField)
   return PQftype(::resID, nField)

METHOD PGResult:fmod(nField)
   return PQfmod(::resID, nField)

METHOD PGResult:fsize(nField)
   return PQfsize(::resID, nField)

METHOD PGResult:getValue( nRow, nCol )
   local c := PQgetvalue(::resID,nRow,nCol)
*   return PGDeref(c)
   return c
METHOD PGResult:binaryTuples()
   return PQbinaryTuples(::resID) == 1

METHOD PGResult:getLength( nRow, nCol )
   return PQgetlength(::resID,nRow,nCol)

METHOD PGResult:getIsNULL( nRow, nCol )
   return PQgetisnull(::resID,nRow,nCol) == 1

METHOD PGResult:cmdStatus()
   local c := PQcmdStatus(::resID)
*   return PGDeref(c)
   return (c)
METHOD PGResult:cmdTuples()
   local c := PQcmdTuples(::resID)
*   return PGDeref(c)
   return (c)

METHOD PGResult:oidValue()
   return PQoidValue(::resID)

// NOT THREAD SAFE
*METHOD PGResult:oidStatus()
*   local c := PQoidStatus(::resID)
*   return PGDeref(c)

METHOD PGResult:escapeString(c)
   return ::connID:escapeString(c)

METHOD PGResult:escapeBytea(cFrom)

*  return ::connID:escapeBytea(cFrom)

  local cbResult := 0
  local pStr     := 0
  local cStr
  if ( Valtype( cFrom ) $ "CM" )
     pStr := PQescapeByteaConn(::connID,cFrom,Len(cFrom),@cbResult)
     if( pStr != 0 )
        cStr := PeekStr(pStr,0,cbResult) ; PQfreemem( pStr )
        return cStr
     end
  end
return NIL


METHOD PGResult:unescapeBytea(cText)
*   return ::connID:unescapeBytea(cText)

  local cbResult := 0
  local pStr     := 0
  local cStr
  if ( Valtype( cText ) $ "CM" )
     pStr := PQunescapeBytea(cText,@cbResult)
     if( pStr != 0 )
        cStr := PeekStr(pStr,0,cbResult) ; PQfreemem( pStr )
        return cStr
     end
  end
return NIL


/***************************

DLLFUNCTION PQbackendPID(o)                     USING STDCALL FROM libpq.dll
DLLFUNCTION PQbinaryTuples(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQclear(c)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQcmdStatus(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQcmdTuples(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectdb(c)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectPoll(u)                    USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectStart(c)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQconsumeInput(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQdb(o)                             USING STDCALL FROM libpq.dll
DLLFUNCTION PQerrorMessage(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQescapeBytea(c,n,@i)               USING STDCALL FROM libpq.dll
DLLFUNCTION PQescapeString(@b,c,l)              USING STDCALL FROM libpq.dll
DLLFUNCTION PQexec(o,c)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQexecParams(o,c,n,ct,cv,cl,cf,rf)  USING STDCALL FROM libpq.dll
DLLFUNCTION PQfinish(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQflush(o)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQfmod(o,i)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQfname(i, n)                       USING STDCALL FROM libpq.dll
DLLFUNCTION PQfnumber(o,c)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQfsize(o,i)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQftype(o,i)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetisnull(o,r,c)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetlength(o,r,c)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetResult(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetvalue(o,n,i)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQhost(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQisBusy(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQisnonblocking(o)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQmakeEmptyPGResult(o,i)            USING STDCALL FROM libpq.dll
DLLFUNCTION PQnfields(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQntuples(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQoidStatus(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQoidValue(o)                       USING STDCALL FROM libpq.dll
DLLFUNCTION PQoptions(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQpass(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQport(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQrequestcancel(o)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQreset(o)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQresetPoll(u)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQresetStart(c)                     USING STDCALL FROM libpq.dll
DLLFUNCTION PQresStatus(i)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQresultErrorMessage(o)             USING STDCALL FROM libpq.dll
DLLFUNCTION PQresultStatus(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQsendQuery(o,c)                    USING STDCALL FROM libpq.dll
DLLFUNCTION PQsetdbLogin(c1,c2,c3,c4,c5,c6,c7)  USING STDCALL FROM libpq.dll
DLLFUNCTION PQsetnonblocking(o,l)               USING STDCALL FROM libpq.dll
DLLFUNCTION PQsocket(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQstatus(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQtrace(o,n)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQunescapeBytea(c,n)                USING STDCALL FROM libpq.dll
DLLFUNCTION PQuntrace(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQuser(o)                           USING STDCALL FROM libpq.dll



init proc load_the_libpq_dll()
* Begin
  local e
  if nLoadLibrary("libpq.dll") == 0
     e := Error():New()
     e:severity  := 3
     e:subsystem := "libpq4xb"
     e:description := "Unable to load libpq.dll"
     Eval( ErrorBlock() , e )
  end
return

***************************/

BEGIN STRUCTURE pgNotify
   MEMBER LPSTR     relname STRZ_READ cRelName
   MEMBER int       be_pid
   MEMBER LPSTR     extra
   MEMBER POINTER32 pNext LINK oNext
END STRUCTURE

BEGIN STRUCTURE pqPrintOpt
   MEMBER BYTEBOOL  header
   MEMBER BYTEBOOL  align
   MEMBER BYTEBOOL  standard
   MEMBER BYTEBOOL  html3
   MEMBER BYTEBOOL  expanded
   MEMBER BYTEBOOL  pager
   GWSTALIGNNEXTMEMBER(8)
   MEMBER LPSTR     fieldSep DYNSZ cFieldSep
   MEMBER LPSTR     tableOpt DYNSZ cTableOpt
   MEMBER LPSTR     caption  DYNSZ cCaption
   MEMBER POINTER32 fieldName
   DYNAMIC METHOD SetFieldNames BLOCK {|s,a| _xfree(s:fieldName), s:fieldName := iif(a == NIL,0,Array2ppMarshall(a)) }
   DYNAMIC METHOD GetFieldNames BLOCK {|s| ppMarshall2Array(s:fieldName) }
END STRUCTURE

BEGIN STRUCTURE pqConnInfoOption
   MEMBER LPSTR keyword   STRZ_READ cKeyword
   MEMBER LPSTR envvar    STRZ_READ cEnvVar
   MEMBER LPSTR compiled  STRZ_READ cCompiled
   MEMBER LPSTR val       STRZ_READ cVal
   MEMBER LPSTR label     STRZ_READ cLabel
   MEMBER LPSTR dispchar  STRZ_READ cDispChar
   MEMBER int   dispsize
END STRUCTURE

BEGIN STRUCTURE pqArgBlock_internal_union
   MEMBER POINTER32 ptr  ; GWSTBACK ; MEMBER int  integer
END STRUCTURE
BEGIN STRUCTURE pqArgBlock
   MEMBER int len
   MEMBER int isint ; GWSTBACK ; MEMBER BOOL lIsInt
   MEMBER POINTER32 ptr  ; GWSTBACK
   MEMBER int integer ; GWSTBACK
   MEMBER @ pqArgBlock_internal_union u
END STRUCTURE

function PQsetdb(host,port,opt,tty,dbname) ; return PQsetdbLogin(host,port,opt,tty,dbname, 0 , 0 )

DLL libpq XIMPORT HANDLE     PQconnectStart(LPSTR conninfo)
DLL libpq XIMPORT int        PQconnectPoll(HANDLE conn)
DLL libpq XIMPORT HANDLE     PQconnectdb(LPSTR conninfo)
DLL libpq XIMPORT HANDLE     PQsetdbLogin(LPSTR host,LPSTR port,LPSTR opt,LPSTR tty,LPSTR db,LPSTR login,LPSTR pwd)
DLL libpq XIMPORT void       PQfinish(HANDLE conn)
DLL libpq XIMPORT POINTER32  PQconndefaults()
DLL libpq XIMPORT void       PQconninfoFree(POINTER32 connOptions)
DLL libpq XIMPORT int        PQresetStart(HANDLE conn)
DLL libpq XIMPORT int        PQresetPoll(HANDLE conn)
DLL libpq XIMPORT void       PQreset(HANDLE conn)
DLL libpq XIMPORT HANDLE     PQgetCancel(HANDLE conn)
DLL libpq XIMPORT void       PQfreeCancel(HANDLE cancel)
DLL libpq XIMPORT int        PQcancel(HANDLE cancel, LPSTR errbuf, int errbufsize) ;
                             DEFAULT errbufsize := iif(Valtype(errbuf)=="C",Len(errbuf),0)
DLL libpq XIMPORT int        PQrequestCancel(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQdb(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQuser(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQpass(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQhost(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQport(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQtty(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQoptions(HANDLE conn)
DLL libpq XIMPORT int        PQstatus(HANDLE conn)
DLL libpq XIMPORT int        PQtransactionStatus(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQparameterStatus(HANDLE conn,LPSTR paramName)
DLL libpq XIMPORT int        PQprotocolVersion(HANDLE conn)
DLL libpq XIMPORT int        PQserverVersion(HANDLE conn)
DLL libpq XIMPORT ZSTRING    PQerrorMessage(HANDLE conn)
DLL libpq XIMPORT int        PQsocket(HANDLE conn)
DLL libpq XIMPORT int        PQbackendPID(HANDLE conn)
DLL libpq XIMPORT int        PQconnectionNeedsPassword(HANDLE conn)
DLL libpq XIMPORT int        PQconnectionUsedPassword(HANDLE conn)
DLL libpq XIMPORT int        PQclientEncoding(HANDLE conn)
DLL libpq XIMPORT int        PQsetClientEncoding(HANDLE conn,LPSTR encoding)
DLL libpq XIMPORT POINTER32  PQgetssl(HANDLE conn)
DLL libpq XIMPORT void       PQinitSSL(int do_init)
DLL libpq XIMPORT int        PQsetErrorVerbosity(HANDLE conn,int verbosity)
DLL libpq XIMPORT void       PQtrace(HANDLE conn, POINTER32 debug_port)
DLL libpq XIMPORT void       PQuntrace(HANDLE conn)
DLL libpq XIMPORT DWORD      PQsetNoticeReceiver(HANDLE conn,DWORD pProc, POINTER32 pArg)
DLL libpq XIMPORT DWORD      PQsetNoticeProcessor(HANDLE conn, DWORD pProc,POINTER32 pArg)
DLL libpq XIMPORT DWORD      PQregisterThreadLock(DWORD pNewhandler)
DLL libpq XIMPORT HANDLE     PQexec(HANDLE conn, LPSTR cQuery)
DLL libpq XIMPORT HANDLE     PQexecParams(HANDLE conn,LPSTR command, int nParams, LPLONG paramTypes,;
                                          POINTER32 pValues, LPLONG pLen, LPLONG pFormats,int resultFormat)
DLL libpq XIMPORT  HANDLE    PQprepare(HANDLE conn, LPSTR cStmtName,LPSTR query, int nParams,LPLONG paramTypes)
DLL libpq XIMPORT  HANDLE    PQexecPrepared(HANDLE conn,LPSTR cStmtName,int nParams, POINTER32 pValues,;
                                            LPLONG pLen, LPLONG pFormats,int resultFormat)
DLL libpq XIMPORT  int       PQsendQuery(HANDLE conn, LPSTR query)
DLL libpq XIMPORT  int       PQsendQueryParams( HANDLE conn,LPSTR command, int nParams,LPLONG paramTypes,;
                                                POINTER32 pValues, LPLONG pLen, LPLONG pFormats,int resultFormat)
DLL libpq XIMPORT  int       PQsendPrepare(HANDLE conn, LPSTR pStmtName, LPSTR query, int nParams,LPLONG paramTypes)
DLL libpq XIMPORT  int       PQsendQueryPrepared(HANDLE conn,LPSTR stmtName,int nParams, POINTER32 pValues,;
                                                 LPLONG pLen, LPLONG pFormats, int resultFormat)
DLL libpq XIMPORT  HANDLE    PQgetResult(HANDLE conn)
DLL libpq XIMPORT  int       PQisBusy(HANDLE conn)
DLL libpq XIMPORT  int       PQconsumeInput(HANDLE conn)
DLL libpq XIMPORT  POINTER32 PQnotifies(HANDLE conn)
DLL libpq XIMPORT  int       PQputCopyData(HANDLE conn,LPSTR buffer, int nBytes) ;
                             DEFAULT nBytes := iif(Valtype(buffer)=="C",Len(buffer),0)
DLL libpq XIMPORT  int       PQputCopyEnd(HANDLE conn,LPSTR errormsg)
DLL libpq XIMPORT  int       PQgetCopyData(HANDLE conn,POINTER32 ppBuffer, BOOL async)
DLL libpq XIMPORT  int       PQsetnonblocking(HANDLE conn, int arg)
DLL libpq XIMPORT  BOOL      PQisnonblocking(HANDLE conn)
DLL libpq XIMPORT  BOOL      PQisthreadsafe()
DLL libpq XIMPORT  BOOL      PQflush(HANDLE conn)
DLL libpq XIMPORT  int       PQresultStatus(HANDLE result)
DLL libpq XIMPORT  ZSTRING   PQresStatus(int status)
DLL libpq XIMPORT  ZSTRING   PQresultErrorMessage(HANDLE result )
DLL libpq XIMPORT  ZSTRING   PQresultErrorField(HANDLE result, int fieldcode)
DLL libpq XIMPORT  int       PQntuples(HANDLE result)
DLL libpq XIMPORT  int       PQnfields(HANDLE result)
DLL libpq XIMPORT  int       PQbinaryTuples(HANDLE result)
DLL libpq XIMPORT  ZSTRING   PQfname(HANDLE result,int field_num)
DLL libpq XIMPORT  int       PQfnumber(HANDLE result, LPSTR field_name)
DLL libpq XIMPORT  int       PQftable(HANDLE result, int field_num)
DLL libpq XIMPORT  int       PQftablecol(HANDLE result, int field_num)
DLL libpq XIMPORT  int       PQfformat(HANDLE result, int field_num)
DLL libpq XIMPORT  int       PQftype(HANDLE result, int field_num)
DLL libpq XIMPORT  int       PQfsize(HANDLE result, int field_num)
DLL libpq XIMPORT  int       PQfmod(HANDLE result, int field_num)
DLL libpq XIMPORT  ZSTRING   PQcmdStatus(HANDLE result)
DLL libpq XIMPORT  INT       PQoidValue(HANDLE result)
DLL libpq XIMPORT  ZSTRING   PQcmdTuples(HANDLE result)
DLL libpq XIMPORT  ZSTRING   PQgetvalue(HANDLE result, int tup_num, int field_num)
DLL libpq XIMPORT  int       PQgetlength(HANDLE result, int tup_num, int field_num)
DLL libpq XIMPORT  int       PQgetisnull(HANDLE result, int tup_num, int field_num)
DLL libpq XIMPORT  int       PQnparams(HANDLE result)
DLL libpq XIMPORT  int       PQparamtype(HANDLE result, int param_num)
DLL libpq XIMPORT  HANDLE    PQdescribePrepared(HANDLE conn, LPSTR stmt)
DLL libpq XIMPORT  HANDLE    PQdescribePortal(HANDLE conn,LPSTR portal)
DLL libpq XIMPORT  int       PQsendDescribePrepared(HANDLE conn,LPSTR stmt)
DLL libpq XIMPORT  int       PQsendDescribePortal(HANDLE conn,LPSTR portal)
DLL libpq XIMPORT  void      PQclear(HANDLE pgResult)
DLL libpq XIMPORT  void      PQfreemem(POINTER32 ptr)
DLL libpq XIMPORT  void      PQfreeNotify(POINTER32 ptr) SYMBOL PQfreemem
DLL libpq XIMPORT  HANDLE    PQmakeEmptyPGresult(HANDLE conn, int status)
DLL libpq XIMPORT  int       PQescapeStringConn(HANDLE conn, LPSTR pTo,LPSTR pFrom,int nLen,LPLONG pError)

DLL libpq XIMPORT  LPBYTE    PQescapeByteaConn(HANDLE conn,LPSTR pFrom, DWORD cbFrom,LPDWORD cbResult)

function cPqEscapeByteA(conn,cFrom)
* Begin
  local cbResult := 0
  local pStr     := 0
  local cStr
  if ( Valtype( cFrom ) $ "CM" )
     pStr := PQescapeByteaConn(conn,cFrom,Len(cFrom),@cbResult)
     if( pStr != 0 )
        cStr := PeekStr(pStr,0,cbResult) ; PQfreemem( pStr )
        return cStr
     end
  end
return NIL

DLL libpq XIMPORT  LPBYTE    PQunescapeBytea(LPSTR pText, LPDWORD cbResult)

function cPqUnescapeByteA(cText)
* Begin
  local cbResult := 0
  local pStr     := 0
  local cStr
  if ( Valtype( cText ) $ "CM" )
     pStr := PQunescapeBytea(cText,@cbResult)
     if( pStr != 0 )
        cStr := PeekStr(pStr,0,cbResult) ; PQfreemem( pStr )
        return cStr
     end
  end
return NIL

DLL libpq XIMPORT  int       lo_open(HANDLE conn, int nObjOid, int mode)
DLL libpq XIMPORT  int       lo_close(HANDLE conn, int fd)
DLL libpq XIMPORT  int       lo_read(HANDLE conn, int fd, LPSTR buf, DWORD nLen)
DLL libpq XIMPORT  int       lo_write(HANDLE conn, int fd,LPSTR buf, DWORD nLen)
DLL libpq XIMPORT  int       lo_lseek(HANDLE conn, int fd, int offset, int whence)
DLL libpq XIMPORT  int       lo_creat(HANDLE conn, int mode)
DLL libpq XIMPORT  int       lo_create(HANDLE conn,int nObjOid)
DLL libpq XIMPORT  int       lo_tell(HANDLE conn, int fd)
DLL libpq XIMPORT  int       lo_truncate(HANDLE conn, int fd,DWORD len)
DLL libpq XIMPORT  int       lo_unlink(HANDLE conn, int nObjOid)
DLL libpq XIMPORT  int       lo_import(HANDLE conn, LPSTR filename)
DLL libpq XIMPORT  int       lo_export(HANDLE conn,int nObjOid, LPSTR filename)
DLL libpq XIMPORT  int       PQmblen(LPSTR s, int encoding)
DLL libpq XIMPORT  int       PQdsplen(LPSTR s, int encoding)
DLL libpq XIMPORT  int       PQenv2encoding()

DLL libpq XIMPORT  LPSTR     PQencryptPassword(LPSTR passwd,LPSTR user)
function cPQencryptPassword(cPwd,cUser)
* Begin
  local pStr := PQencryptPassword(cPwd,cUser)
  local cStr
  if ( pStr != 0 )
     cStr := PeekStr(pStr,0,-1); PQfreemem( pStr )
     return cStr
  end
return NIL

DLL libpq XIMPORT  int       pg_char_to_encoding(LPSTR name)
DLL libpq XIMPORT  ZSTRING   pg_encoding_to_char(int encoding)
DLL libpq XIMPORT  int       pg_valid_server_encoding_id(int encoding)

//-------------------------------------------------------------------------------------------------------------------------
// Deprecated routines for copy in/out
// DLL libpq XIMPORT  int PQgetline(HANDLE conn,LPSTR string, int len)
// DLL libpq XIMPORT  int PQputline(HANDLE conn,LPSTR string)
// DLL libpq XIMPORT  int PQgetlineAsync(HANDLE conn,LPSTR buffer, int bufsize);
// DEFAULT bufsize := iif(Valtype(buffer)=="C",Len(buffer),0)
// DLL libpq XIMPORT  int PQputnbytes(HANDLE conn,LPSTR buffer, int nbytes)
// DEFAULT nbytes := iif(Valtype(buffer)=="C",Len(buffer),0)
// DLL libpq XIMPORT  int PQendcopy(HANDLE conn)
//-------------------------------------------------------------------------------------------------------------------------
// "Fast path" interface --- not really recommended for application use
// DLL libpq XIMPORT  PGresult *PQfn( HANDLE conn,int fnid, LPLONG result_buf, LPLONG result_len,;
//                                    BOOL result_is_int, POINTER32 args, int nargs)
//-------------------------------------------------------------------------------------------------------------------------
// PQoidStatus is deprecated in favor of PQoidValue. It is not thread-safe.
// DLL libpq XIMPORT  ZSTRING   PQoidStatus(HANDLE result) // NOT THREAD SAFE
//-------------------------------------------------------------------------------------------------------------------------
// These forms are deprecated
// DLL libpq XIMPORT  DWORD     PQescapeString(LPSTR pTo,LPSTR pFrom, int nLen)
// DLL libpq XIMPORT  LPBYTE    PQescapeBytea(LPSTR pFrom, DWORD cbFrom,LPDWORD pcbTo)
//-------------------------------------------------------------------------------------------------------------------------
// The contents of FILE structure vary on single/multi thread and debug on/off compilation
// DLL libpq XIMPORT  void      PQprint(POINTER32 pFILE_out,HANDLE result,POINTER32 ps)
//  really old printing routines
// DLL libpq XIMPORT  void      PQdisplayTuples(HANDLE result,POINTER32 pFILE,BOOL nFillAlign,;
//                                              LPSTR cFieldSep, BOOL bPrintHeader,BOOL bQuiet)
// DLL libpq XIMPORT  void      PQprintTuples(HANDLE result,POINTER32 pFILE, BOOL bPrintAttName,;
//                                            int terseOutput, int width)
//-------------------------------------------------------------------------------------------------------------------------